import datetime
from typing import Dict, List, Optional, Tuple, Union

import dash
import pandas as pd
from dash import Input, Output, State, callback
from dash.exceptions import PreventUpdate
from webviz_config import EncodedFile, WebvizPluginABC
from webviz_config._theme_class import WebvizConfigTheme
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import ViewABC
from webviz_subsurface_components import ExpressionInfo, VectorDefinition

from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)
from webviz_subsurface._utils.unique_theming import unique_colors
from webviz_subsurface._utils.vector_calculator import get_selected_expressions

from ._property_serialization import (
    EnsembleSubplotBuilder,
    GraphFigureBuilderBase,
    VectorSubplotBuilder,
)
from ._settings._ensembles import EnsemblesSettings
from ._settings._filter_realization import FilterRealizationSettings
from ._settings._group_by import GroupBySettings
from ._settings._resampling_frequency import ResamplingFrequencySettings
from ._settings._time_series import TimeSeriesSettings
from ._settings._visualization import VisualizationSettings
from ._types import (
    DeltaEnsemble,
    FanchartOptions,
    StatisticsFromOptions,
    StatisticsOptions,
    SubplotGroupByOptions,
    TraceOptions,
    VisualizationOptions,
)
from ._utils import DerivedVectorsAccessor, datetime_utils
from ._utils.derived_ensemble_vectors_accessor_utils import (
    create_derived_vectors_accessor_dict,
)
from ._utils.ensemble_summary_provider_set_utils import (
    create_vector_plot_titles_from_provider_set,
)
from ._utils.from_timeseries_cumulatives import (
    datetime_to_intervalstr,
    is_per_interval_or_per_day_vector,
)
from ._utils.history_vectors import create_history_vectors_df
from ._utils.trace_line_shape import get_simulation_line_shape
from ._utils.vector_statistics import create_vectors_statistics_df
from ._view_elements import SubplotGraph


class SubplotView(ViewABC):
    class Ids(StrEnum):
        SUBPLOT = "subplot"

        ENSEMBLE_SETTINGS = "ensemble-settings"
        FILTER_REALIZATION_SETTINGS = "filter-realization-settings"
        GROUP_BY_SETTINGS = "group-by-settings"
        RESAMPLING_FREQUENCY_SETTINGS = "resampling-frequency-settings"
        TIME_SERIES_SETTINGS = "time-series-settings"
        VISUALIZATION_SETTINGS = "visualization-settings"

    # pylint: disable=too-many-arguments, too-many-locals
    def __init__(
        self,
        custom_vector_definitions: dict,
        custom_vector_definitions_base: dict,
        disable_resampling_dropdown: bool,
        initial_selected_vectors: List[str],
        initial_vector_selector_data: list,
        initial_visualization: VisualizationOptions,
        input_provider_set: EnsembleSummaryProviderSet,
        predefined_expressions: List[ExpressionInfo],
        selected_resampling_frequency: Frequency,
        vector_calculator_data: List,
        vector_selector_base_data: list,
        theme: WebvizConfigTheme,
        user_defined_vector_definitions: Dict[str, VectorDefinition],
        observations: dict,  # TODO: Improve typehint?
        line_shape_fallback: str = "linear",
    ) -> None:
        super().__init__("Subplot View")

        column = self.add_column()
        column.add_view_element(SubplotGraph(), SubplotView.Ids.SUBPLOT)

        self.add_settings_groups(
            {
                SubplotView.Ids.GROUP_BY_SETTINGS: GroupBySettings(),
                SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS: ResamplingFrequencySettings(
                    disable_resampling_dropdown=disable_resampling_dropdown,
                    selected_resampling_frequency=selected_resampling_frequency,
                    ensembles_dates=input_provider_set.all_dates(
                        selected_resampling_frequency
                    ),
                    input_provider_set=input_provider_set,
                ),
                SubplotView.Ids.ENSEMBLE_SETTINGS: EnsemblesSettings(
                    ensembles_names=input_provider_set.provider_names(),
                    input_provider_set=input_provider_set,
                ),
                SubplotView.Ids.TIME_SERIES_SETTINGS: TimeSeriesSettings(
                    initial_vector_selector_data=initial_vector_selector_data,
                    custom_vector_definitions=custom_vector_definitions,
                    vector_calculator_data=vector_calculator_data,
                    predefined_expressions=predefined_expressions,
                    vector_selector_base_data=vector_selector_base_data,
                    custom_vector_definitions_base=custom_vector_definitions_base,
                    initial_selected_vectors=initial_selected_vectors,
                ),
                SubplotView.Ids.VISUALIZATION_SETTINGS: VisualizationSettings(
                    selected_visualization=initial_visualization
                ),
                SubplotView.Ids.FILTER_REALIZATION_SETTINGS: FilterRealizationSettings(
                    realizations=input_provider_set.all_realizations()
                ),
            }
        )

        self._initial_selected_vectors = initial_selected_vectors
        self._input_provider_set = input_provider_set
        self._theme = theme
        self._line_shape_fallback = line_shape_fallback
        self._user_defined_vector_definitions = user_defined_vector_definitions
        self._observations = observations

    # pylint: disable=too-many-statements
    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.view_element_unique_id(
                    SubplotView.Ids.SUBPLOT, SubplotGraph.Ids.GRAPH
                ),
                "figure",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.TIME_SERIES_SETTINGS,
                    TimeSeriesSettings.Ids.VECTOR_SELECTOR,
                ),
                "selectedNodes",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.ENSEMBLE_SETTINGS,
                    EnsemblesSettings.Ids.ENSEMBLES_DROPDOWN,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.VISUALIZATION_SETTINGS,
                    VisualizationSettings.Ids.VISUALIZATION_RADIO_ITEMS,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.VISUALIZATION_SETTINGS,
                    VisualizationSettings.Ids.PLOT_STATISTICS_OPTIONS_CHECKLIST,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.VISUALIZATION_SETTINGS,
                    VisualizationSettings.Ids.PLOT_FANCHART_OPTIONS_CHECKLIST,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.VISUALIZATION_SETTINGS,
                    VisualizationSettings.Ids.PLOT_TRACE_OPTIONS_CHECKLIST,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.GROUP_BY_SETTINGS,
                    GroupBySettings.Ids.SUBPLOT_OWNER_OPTIONS_RADIO_ITEMS,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS,
                    ResamplingFrequencySettings.Ids.RESAMPLING_FREQUENCY_DROPDOWN,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.FILTER_REALIZATION_SETTINGS,
                    FilterRealizationSettings.Ids.REALIZATIONS_FILTER_SELECTOR,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.FILTER_REALIZATION_SETTINGS,
                    FilterRealizationSettings.Ids.STATISTICS_FROM_RADIO_ITEMS,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS,
                    ResamplingFrequencySettings.Ids.RELATIVE_DATE_DROPDOWN,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.TIME_SERIES_SETTINGS,
                    TimeSeriesSettings.Ids.GRAPH_DATA_HAS_CHANGED_TRIGGER,
                ),
                "data",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.ENSEMBLE_SETTINGS,
                    EnsemblesSettings.Ids.CREATED_DELTA_ENSEMBLES_STORE,
                ),
                "data",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.TIME_SERIES_SETTINGS,
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_EXPRESSIONS,
                ),
                "data",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.ENSEMBLE_SETTINGS,
                    EnsemblesSettings.Ids.ENSEMBLES_DROPDOWN,
                ),
                "options",
            ),
        )
        @callback_typecheck
        # pylint: disable=too-many-arguments, too-many-locals, too-many-branches, too-many-statements
        def _update_graph(
            vectors: List[str],
            selected_ensembles: List[str],
            visualization: VisualizationOptions,
            statistics_options: List[StatisticsOptions],
            fanchart_options: List[FanchartOptions],
            trace_options: List[TraceOptions],
            subplot_group_by: SubplotGroupByOptions,
            resampling_frequency_value: str,
            selected_realizations: List[int],
            statistics_from: StatisticsFromOptions,
            relative_date_value: Optional[str],
            _graph_data_has_changed_trigger: int,
            delta_ensembles: List[DeltaEnsemble],
            vector_calculator_expressions: List[ExpressionInfo],
            ensemble_dropdown_options: List[dict],
        ) -> dict:
            """Callback to update all graphs based on selections

            * De-serialize from JSON serializable format to strongly typed and filtered format
            * Business logic:
                * Functionality with "strongly typed" and filtered input format - functions and
                classes
                * EnsembleSummaryProviderSet for EnsembleSummaryProviders, i.e. input_provider_set
                * DerivedEnsembleVectorsAccessor to access derived vector data from ensembles
                with single providers or delta ensemble with two providers
                * GraphFigureBuilder to create graph with subplots per vector or subplots per
                ensemble, using VectorSubplotBuilder and EnsembleSubplotBuilder, respectively
            * Create/build property serialization in FigureBuilder by use of business logic data

            NOTE: _graph_data_has_changed_trigger is only used to trigger callback when change of
            graphs data has changed and re-render of graph is necessary. E.g. when a selected
            expression from the VectorCalculator gets edited without changing the expression
            name - i.e.
            VectorSelector selectedNodes remain unchanged.
            """
            if not isinstance(selected_ensembles, list):
                raise TypeError("ensembles should always be of type list")

            if vectors is None:
                vectors = self._initial_selected_vectors

            # Retrieve the selected expressions
            selected_expressions = get_selected_expressions(
                vector_calculator_expressions, vectors
            )

            # Convert from string values to strongly typed
            resampling_frequency = Frequency.from_string_value(
                resampling_frequency_value
            )

            all_ensemble_names = [
                option["value"] for option in ensemble_dropdown_options
            ]

            relative_date: Optional[datetime.datetime] = (
                None
                if relative_date_value is None
                else datetime_utils.from_str(relative_date_value)
            )

            # Prevent update if realization filtering is not affecting pure statistics plot
            # TODO: Refactor code or create utility for getting trigger ID in a "cleaner" way?
            ctx = dash.callback_context.triggered
            trigger_id = ctx[0]["prop_id"].split(".")[0]

            if (
                trigger_id
                == self.settings_group_unique_id(
                    SubplotView.Ids.FILTER_REALIZATION_SETTINGS,
                    FilterRealizationSettings.Ids.REALIZATIONS_FILTER_SELECTOR,
                )
                and statistics_from is StatisticsFromOptions.ALL_REALIZATIONS
                and visualization
                in [
                    VisualizationOptions.STATISTICS,
                    VisualizationOptions.FANCHART,
                ]
            ):
                raise PreventUpdate

            # Create dict of derived vectors accessors for selected ensembles
            derived_vectors_accessors: Dict[
                str, DerivedVectorsAccessor
            ] = create_derived_vectors_accessor_dict(
                ensembles=selected_ensembles,
                vectors=vectors,
                provider_set=self._input_provider_set,
                expressions=selected_expressions,
                delta_ensembles=delta_ensembles,
                resampling_frequency=resampling_frequency,
                relative_date=relative_date,
            )

            # TODO: How to get metadata for calculated vector?
            vector_line_shapes: Dict[str, str] = {
                vector: get_simulation_line_shape(
                    self._line_shape_fallback,
                    vector,
                    self._input_provider_set.vector_metadata(vector),
                )
                for vector in vectors
            }

            figure_builder: GraphFigureBuilderBase
            if subplot_group_by is SubplotGroupByOptions.VECTOR:
                # Create unique colors based on all ensemble names to preserve consistent colors
                ensemble_colors = unique_colors(all_ensemble_names, self._theme)
                vector_titles = create_vector_plot_titles_from_provider_set(
                    vectors,
                    selected_expressions,
                    self._input_provider_set,
                    self._user_defined_vector_definitions,
                    resampling_frequency,
                )
                figure_builder = VectorSubplotBuilder(
                    vectors,
                    vector_titles,
                    ensemble_colors,
                    resampling_frequency,
                    vector_line_shapes,
                    self._theme,
                )
            elif subplot_group_by is SubplotGroupByOptions.ENSEMBLE:
                vector_colors = unique_colors(vectors, self._theme)
                figure_builder = EnsembleSubplotBuilder(
                    vectors,
                    selected_ensembles,
                    vector_colors,
                    resampling_frequency,
                    vector_line_shapes,
                    self._theme,
                )
            else:
                raise PreventUpdate

            # Get all realizations if statistics accross all realizations are requested
            is_statistics_from_all_realizations = (
                statistics_from == StatisticsFromOptions.ALL_REALIZATIONS
                and visualization
                in [
                    VisualizationOptions.FANCHART,
                    VisualizationOptions.STATISTICS,
                    VisualizationOptions.STATISTICS_AND_REALIZATIONS,
                ]
            )

            # Plotting per derived vectors accessor
            for ensemble, accessor in derived_vectors_accessors.items():
                # Realization query - realizations query for accessor
                # - Get non-filter query, None, if statistics from all realizations is needed
                # - Create valid realizations query for accessor otherwise:
                #   * List[int]: Filtered valid realizations, empty list if none are valid
                #   * None: Get all realizations, i.e. non-filtered query
                realizations_query = (
                    None
                    if is_statistics_from_all_realizations
                    else accessor.create_valid_realizations_query(selected_realizations)
                )

                # If all selected realizations are invalid for accessor - empty list
                if realizations_query == []:
                    continue

                # TODO: Consider to remove list vectors_df_list and use pd.concat to obtain
                # one single dataframe with vector columns. NB: Assumes equal sampling rate
                # for each vector type - i.e equal number of rows in dataframes

                # Retrive vectors data from accessor
                vectors_df_list: List[pd.DataFrame] = []
                if accessor.has_provider_vectors():
                    vectors_df_list.append(
                        accessor.get_provider_vectors_df(
                            realizations=realizations_query
                        )
                    )
                if accessor.has_per_interval_and_per_day_vectors():
                    vectors_df_list.append(
                        accessor.create_per_interval_and_per_day_vectors_df(
                            realizations=realizations_query
                        )
                    )
                if accessor.has_vector_calculator_expressions():
                    vectors_df_list.append(
                        accessor.create_calculated_vectors_df(
                            realizations=realizations_query
                        )
                    )

                for vectors_df in vectors_df_list:
                    # Ensure rows of data
                    if not vectors_df.shape[0]:
                        continue

                    if visualization == VisualizationOptions.REALIZATIONS:
                        # Show selected realizations - only filter df if realizations filter
                        # query is not performed
                        figure_builder.add_realizations_traces(
                            vectors_df
                            if realizations_query
                            else vectors_df[
                                vectors_df["REAL"].isin(selected_realizations)
                            ],
                            ensemble,
                        )
                    if visualization == VisualizationOptions.STATISTICS:
                        vectors_statistics_df = create_vectors_statistics_df(vectors_df)
                        figure_builder.add_statistics_traces(
                            vectors_statistics_df,
                            ensemble,
                            statistics_options,
                        )
                    if visualization == VisualizationOptions.FANCHART:
                        vectors_statistics_df = create_vectors_statistics_df(vectors_df)
                        figure_builder.add_fanchart_traces(
                            vectors_statistics_df,
                            ensemble,
                            fanchart_options,
                        )
                    if (
                        visualization
                        == VisualizationOptions.STATISTICS_AND_REALIZATIONS
                    ):
                        # Configure line width and color scaling to easier separate
                        # statistics traces and realization traces.
                        # Show selected realizations - only filter df if realizations filter
                        # query is not performed
                        figure_builder.add_realizations_traces(
                            vectors_df
                            if realizations_query
                            else vectors_df[
                                vectors_df["REAL"].isin(selected_realizations)
                            ],
                            ensemble,
                            color_lightness_scale=150.0,
                        )
                        # Add statistics on top
                        vectors_statistics_df = create_vectors_statistics_df(vectors_df)
                        figure_builder.add_statistics_traces(
                            vectors_statistics_df,
                            ensemble,
                            statistics_options,
                            line_width=3,
                        )

            # Retrieve selected input providers
            selected_input_providers = EnsembleSummaryProviderSet(
                {
                    name: provider
                    for name, provider in self._input_provider_set.items()
                    if name in selected_ensembles
                }
            )

            # Do not add observations if only delta ensembles are selected
            is_only_delta_ensembles = (
                len(selected_input_providers.provider_names()) == 0
                and len(derived_vectors_accessors) > 0
            )
            if (
                self._observations
                and TraceOptions.OBSERVATIONS in trace_options
                and not is_only_delta_ensembles
                and not relative_date
            ):
                for vector in vectors:
                    vector_observations = self._observations.get(vector)
                    if vector_observations:
                        figure_builder.add_vector_observations(
                            vector, vector_observations
                        )

            # Add history trace
            # TODO: Improve when new history vector input format is in place
            if TraceOptions.HISTORY in trace_options and not relative_date:
                if (
                    isinstance(figure_builder, VectorSubplotBuilder)
                    and len(selected_input_providers.provider_names()) > 0
                ):
                    # Add history trace using first selected ensemble
                    name = selected_input_providers.provider_names()[0]
                    provider = selected_input_providers.provider(name)
                    vector_names = provider.vector_names()

                    provider_vectors = [elm for elm in vectors if elm in vector_names]
                    if provider_vectors:
                        history_vectors_df = create_history_vectors_df(
                            provider, provider_vectors, resampling_frequency
                        )
                        # TODO: Handle check of non-empty dataframe better?
                        if (
                            not history_vectors_df.empty
                            and "DATE" in history_vectors_df.columns
                        ):
                            figure_builder.add_history_traces(history_vectors_df)

                if isinstance(figure_builder, EnsembleSubplotBuilder):
                    # Add history trace for each ensemble
                    for name, provider in selected_input_providers.items():
                        vector_names = provider.vector_names()

                        provider_vectors = [
                            elm for elm in vectors if elm in vector_names
                        ]
                        if provider_vectors:
                            history_vectors_df = create_history_vectors_df(
                                provider, provider_vectors, resampling_frequency
                            )
                            # TODO: Handle check of non-empty dataframe better?
                            if (
                                not history_vectors_df.empty
                                and "DATE" in history_vectors_df.columns
                            ):
                                figure_builder.add_history_traces(
                                    history_vectors_df,
                                    name,
                                )

            # Create legends when all data is added to figure
            figure_builder.create_graph_legends()

            return figure_builder.get_serialized_figure()

        @callback(
            self.view_data_output(),
            self.view_data_requested(),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.TIME_SERIES_SETTINGS,
                    TimeSeriesSettings.Ids.VECTOR_SELECTOR,
                ),
                "selectedNodes",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.ENSEMBLE_SETTINGS,
                    EnsemblesSettings.Ids.ENSEMBLES_DROPDOWN,
                ),
                "value",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.VISUALIZATION_SETTINGS,
                    VisualizationSettings.Ids.VISUALIZATION_RADIO_ITEMS,
                ),
                "value",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS,
                    ResamplingFrequencySettings.Ids.RESAMPLING_FREQUENCY_DROPDOWN,
                ),
                "value",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.FILTER_REALIZATION_SETTINGS,
                    FilterRealizationSettings.Ids.REALIZATIONS_FILTER_SELECTOR,
                ),
                "value",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.FILTER_REALIZATION_SETTINGS,
                    FilterRealizationSettings.Ids.STATISTICS_FROM_RADIO_ITEMS,
                ),
                "value",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.ENSEMBLE_SETTINGS,
                    EnsemblesSettings.Ids.CREATED_DELTA_ENSEMBLES_STORE,
                ),
                "data",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.TIME_SERIES_SETTINGS,
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_EXPRESSIONS,
                ),
                "data",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS,
                    ResamplingFrequencySettings.Ids.RELATIVE_DATE_DROPDOWN,
                ),
                "value",
            ),
        )
        @callback_typecheck
        # pylint: disable=too-many-branches,
        def _user_download_data(
            data_requested: Union[int, None],
            vectors: List[str],
            selected_ensembles: List[str],
            visualization: VisualizationOptions,
            resampling_frequency_value: str,
            selected_realizations: List[int],
            statistics_from: StatisticsFromOptions,
            delta_ensembles: List[DeltaEnsemble],
            vector_calculator_expressions: List[ExpressionInfo],
            relative_date_value: Optional[str],
        ) -> Union[EncodedFile, str]:
            """Callback to download data based on selections

            Retrieve vector data based on selected visualizations and filtered realizations

            NOTE:
            * Does not group based on "Group By" - data is stored per vector
            * All statistics included - no filtering on statistics selections
            * No history vector
            * No observation data
            """
            if data_requested is None:
                raise PreventUpdate

            if not isinstance(selected_ensembles, list):
                raise TypeError("ensembles should always be of type list")

            if vectors is None:
                vectors = self._initial_selected_vectors

            # Retrieve the selected expressions
            selected_expressions = get_selected_expressions(
                vector_calculator_expressions, vectors
            )

            # Convert from string values to strongly typed
            resampling_frequency = Frequency.from_string_value(
                resampling_frequency_value
            )
            relative_date: Optional[datetime.datetime] = (
                None
                if relative_date_value is None
                else datetime_utils.from_str(relative_date_value)
            )

            # Create dict of derived vectors accessors for selected ensembles
            derived_vectors_accessors: Dict[
                str, DerivedVectorsAccessor
            ] = create_derived_vectors_accessor_dict(
                ensembles=selected_ensembles,
                vectors=vectors,
                provider_set=self._input_provider_set,
                expressions=selected_expressions,
                delta_ensembles=delta_ensembles,
                resampling_frequency=resampling_frequency,
                relative_date=relative_date,
            )

            # Dict with vector name as key and dataframe data as value
            vector_dataframe_dict: Dict[str, pd.DataFrame] = {}

            # Get all realizations if statistics across all realizations are requested
            is_statistics_from_all_realizations = (
                statistics_from == StatisticsFromOptions.ALL_REALIZATIONS
                and visualization
                in [
                    VisualizationOptions.FANCHART,
                    VisualizationOptions.STATISTICS,
                    VisualizationOptions.STATISTICS_AND_REALIZATIONS,
                ]
            )

            # Plotting per derived vectors accessor
            for ensemble, accessor in derived_vectors_accessors.items():
                # Realization query - realizations query for accessor
                # - Get non-filter query, None, if statistics from all realizations is needed
                # - Create valid realizations query for accessor otherwise:
                #   * List[int]: Filtered valid realizations, empty list if none are valid
                #   * None: Get all realizations, i.e. non-filtered query
                realizations_query = (
                    None
                    if is_statistics_from_all_realizations
                    else accessor.create_valid_realizations_query(selected_realizations)
                )

                # If all selected realizations are invalid for accessor - empty list
                if realizations_query == []:
                    continue

                # Retrieve vectors data from accessor
                vectors_df_list: List[pd.DataFrame] = []
                if accessor.has_provider_vectors():
                    vectors_df_list.append(
                        accessor.get_provider_vectors_df(
                            realizations=realizations_query
                        )
                    )
                if accessor.has_per_interval_and_per_day_vectors():
                    vectors_df_list.append(
                        accessor.create_per_interval_and_per_day_vectors_df(
                            realizations=realizations_query
                        )
                    )
                if accessor.has_vector_calculator_expressions():
                    vectors_df_list.append(
                        accessor.create_calculated_vectors_df(
                            realizations=realizations_query
                        )
                    )

                # Append data for each vector
                for vectors_df in vectors_df_list:
                    # Ensure rows of data
                    if not vectors_df.shape[0]:
                        continue

                    vector_names = [
                        elm for elm in vectors_df.columns if elm not in ["DATE", "REAL"]
                    ]

                    if visualization in [
                        VisualizationOptions.REALIZATIONS,
                        VisualizationOptions.STATISTICS_AND_REALIZATIONS,
                    ]:
                        # NOTE: Should in theory not have situation with query of all realizations
                        # if not wanted
                        vectors_df_filtered = (
                            vectors_df
                            if realizations_query
                            else vectors_df[
                                vectors_df["REAL"].isin(selected_realizations)
                            ]
                        )
                        for vector in vector_names:
                            vector_df = vectors_df_filtered[["DATE", "REAL", vector]]
                            row_count = vector_df.shape[0]
                            ensemble_name_list = [ensemble] * row_count
                            vector_df.insert(
                                loc=0, column="ENSEMBLE", value=ensemble_name_list
                            )

                            if is_per_interval_or_per_day_vector(vector):
                                vector_df["DATE"] = vector_df["DATE"].apply(
                                    datetime_to_intervalstr, freq=resampling_frequency
                                )

                            vector_key = vector + "_realizations"
                            if vector_dataframe_dict.get(vector_key) is None:
                                vector_dataframe_dict[vector_key] = vector_df
                            else:
                                vector_dataframe_dict[vector_key] = pd.concat(
                                    [vector_dataframe_dict[vector_key], vector_df],
                                    ignore_index=True,
                                    axis=0,
                                )

                    if visualization in [
                        VisualizationOptions.STATISTICS,
                        VisualizationOptions.FANCHART,
                        VisualizationOptions.STATISTICS_AND_REALIZATIONS,
                    ]:
                        vectors_statistics_df = create_vectors_statistics_df(vectors_df)

                        for vector in vector_names:
                            vector_statistics_df = vectors_statistics_df[
                                ["DATE", vector]
                            ]
                            row_count = vector_statistics_df.shape[0]
                            ensemble_name_list = [ensemble] * row_count
                            vector_statistics_df.insert(
                                loc=0, column="ENSEMBLE", value=ensemble_name_list
                            )

                            vector_key = vector + "_statistics"

                            if is_per_interval_or_per_day_vector(vector):
                                # Copy df to prevent SettingWithCopyWarning
                                vector_statistics_df = vector_statistics_df.copy()
                                vector_statistics_df["DATE"] = vector_statistics_df[
                                    "DATE"
                                ].apply(
                                    datetime_to_intervalstr, freq=resampling_frequency
                                )
                            if vector_dataframe_dict.get(vector_key) is None:
                                vector_dataframe_dict[vector_key] = vector_statistics_df
                            else:
                                vector_dataframe_dict[vector_key] = pd.concat(
                                    [
                                        vector_dataframe_dict[vector_key],
                                        vector_statistics_df,
                                    ],
                                    ignore_index=True,
                                    axis=0,
                                )

            # : is replaced with _ in filenames to stay within POSIX portable pathnames
            # (e.g. : is not valid in a Windows path)
            return WebvizPluginABC.plugin_data_compress(
                [
                    {
                        "filename": f"{vector.replace(':', '_')}.csv",
                        "content": df.to_csv(index=False),
                    }
                    for vector, df in vector_dataframe_dict.items()
                ]
            )

        @callback(
            Output(
                self.settings_group_unique_id(
                    SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS,
                    ResamplingFrequencySettings.Ids.RELATIVE_DATE_DROPDOWN,
                ),
                "options",
            ),
            Output(
                self.settings_group_unique_id(
                    SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS,
                    ResamplingFrequencySettings.Ids.RELATIVE_DATE_DROPDOWN,
                ),
                "value",
            ),
            Output(
                self.settings_group_unique_id(
                    SubplotView.Ids.VISUALIZATION_SETTINGS,
                    VisualizationSettings.Ids.PLOT_TRACE_OPTIONS_CHECKLIST,
                ),
                "style",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS,
                    ResamplingFrequencySettings.Ids.RESAMPLING_FREQUENCY_DROPDOWN,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS,
                    ResamplingFrequencySettings.Ids.RELATIVE_DATE_DROPDOWN,
                ),
                "value",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS,
                    ResamplingFrequencySettings.Ids.RELATIVE_DATE_DROPDOWN,
                ),
                "options",
            ),
            State(
                self.settings_group_unique_id(
                    SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS,
                    ResamplingFrequencySettings.Ids.RELATIVE_DATE_DROPDOWN,
                ),
                "value",
            ),
        )
        @callback_typecheck
        def _update_relative_date_dropdown_and_trace_options_style(
            resampling_frequency_value: str,
            relative_date_value: Optional[str],
            current_relative_date_options: List[Dict[str, str]],
            current_relative_date_value: Optional[str],
        ) -> Tuple[List[Dict[str, str]], Optional[str], Dict[str, str]]:
            """This callback updates dropdown based on selected resampling frequency selection
            and hide trace options (History and Observation) when a relative date is selected.

            If dates are not existing for a provider, the data accessor must handle invalid
            relative date selection!
            """
            resampling_frequency = Frequency.from_string_value(
                resampling_frequency_value
            )
            dates_union = self._input_provider_set.all_dates(resampling_frequency)

            # Create dropdown options:
            new_relative_date_options: List[Dict[str, str]] = [
                {
                    "label": datetime_utils.to_str(_date),
                    "value": datetime_utils.to_str(_date),
                }
                for _date in dates_union
            ]

            # Create valid dropdown value:
            new_relative_date_value = next(
                (
                    elm["value"]
                    for elm in new_relative_date_options
                    if elm["value"] == current_relative_date_value
                ),
                None,
            )

            # Prevent updates if unchanged
            if new_relative_date_options == current_relative_date_options:
                new_relative_date_options = dash.no_update
            if new_relative_date_value == current_relative_date_value:
                new_relative_date_value = dash.no_update

            # Convert to Optional[datetime.datetime]
            relative_date: Optional[datetime.datetime] = (
                None
                if relative_date_value is None
                else datetime_utils.from_str(relative_date_value)
            )
            trace_options_style = (
                {"display": "none"} if relative_date else {"display": "block"}
            )

            return (
                new_relative_date_options,
                new_relative_date_value,
                trace_options_style,
            )



from typing import Dict, List, Union

import webviz_core_components as wcc
from dash import Input, Output, callback
from dash.development.base_component import Component
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC, ViewABC

from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)

from .._plugin_ids import PluginIds
from ..utils import make_dataframes as makedf
from ..utils import make_figures as makefigs
from ._view_functions import _get_well_names_combined


class PlotSettingsMisfit(SettingsGroupABC):
    class Ids:
        # pylint: disable=too-few-public-methods
        COLORBY = "colorby"
        SORTING_RANKING = "sorting-ranking"
        FIG_LAYOUT_HEIGHT = "fig-layout-height"

    def __init__(self) -> None:
        super().__init__("Plot Settings")

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                label="Color by",
                id=self.register_component_unique_id(PlotSettingsMisfit.Ids.COLORBY),
                options=[
                    {
                        "label": "Total misfit",
                        "value": "misfit",
                    },
                    {"label": "Phases", "value": "phases"},
                    {"label": "Date", "value": "date"},
                    {"label": "None", "value": None},
                ],
                value="phases",
                multi=False,
                clearable=False,
                persistence=True,
                persistence_type="memory",
            ),
            wcc.Dropdown(
                label="Sorting/ranking",
                id=self.register_component_unique_id(
                    PlotSettingsMisfit.Ids.SORTING_RANKING
                ),
                options=[
                    {
                        "label": "None",
                        "value": None,
                    },
                    {
                        "label": "Ascending",
                        "value": "total ascending",
                    },
                    {
                        "label": "Descending",
                        "value": "total descending",
                    },
                ],
                value="total ascending",
                multi=False,
                clearable=False,
                persistence=True,
                persistence_type="memory",
            ),
            wcc.Dropdown(
                label="Fig layout - height",
                id=self.register_component_unique_id(
                    PlotSettingsMisfit.Ids.FIG_LAYOUT_HEIGHT
                ),
                options=[
                    {
                        "label": "Very small",
                        "value": 250,
                    },
                    {
                        "label": "Small",
                        "value": 350,
                    },
                    {
                        "label": "Medium",
                        "value": 450,
                    },
                    {
                        "label": "Large",
                        "value": 700,
                    },
                    {
                        "label": "Very large",
                        "value": 1000,
                    },
                ],
                value=450,
                clearable=False,
                persistence=True,
                persistence_type="memory",
            ),
        ]


class MisfitOptions(SettingsGroupABC):
    class Ids:
        # pylint: disable=too-few-public-methods
        MISFIT_WEIGHT = "misfit-weight"
        MISFIT_EXPONENT = "misfit-exponent"

    def __init__(self) -> None:
        super().__init__("Misfit options")

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                label="Misfit weight",
                id=self.register_component_unique_id(MisfitOptions.Ids.MISFIT_WEIGHT),
                options=[
                    {
                        "label": "Phase weights",
                        "value": -1.0,
                    },
                    {"label": "None", "value": 0.0},
                    {
                        "label": "10% obs error (min=1000)",
                        "value": 0.10,
                    },
                    {
                        "label": "20% obs error (min=1000)",
                        "value": 0.20,
                    },
                ],
                value=-1.0,
                clearable=False,
                persistence=True,
                persistence_type="memory",
            ),
            wcc.Dropdown(
                label="Misfit exponent",
                id=self.register_component_unique_id(MisfitOptions.Ids.MISFIT_EXPONENT),
                options=[
                    {
                        "label": "Linear sum",
                        "value": 1.0,
                    },
                    {
                        "label": "Squared sum",
                        "value": 2.0,
                    },
                ],
                value=1.0,
                clearable=False,
                persistence=True,
                persistence_type="memory",
            ),
        ]


class MisfitPerRealView(ViewABC):
    class Ids:
        # pylint: disable=too-few-public-methods
        PLOT_SETTINGS = "plot-settings"
        MISFIT_OPTIONS = "misfit-options"
        MAIN_COLUMN = "main-column"

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        input_provider_set: EnsembleSummaryProviderSet,
        ens_vectors: Dict[str, List[str]],
        ens_realizations: Dict[str, List[int]],
        well_collections: Dict[str, List[str]],
        weight_reduction_factor_oil: float,
        weight_reduction_factor_wat: float,
        weight_reduction_factor_gas: float,
    ) -> None:
        super().__init__("Production misfit per real")

        self.input_provider_set = input_provider_set
        self.ens_vectors = ens_vectors
        self.ens_realizations = ens_realizations
        self.well_collections = well_collections
        self.weight_reduction_factor_oil = weight_reduction_factor_oil
        self.weight_reduction_factor_wat = weight_reduction_factor_wat
        self.weight_reduction_factor_gas = weight_reduction_factor_gas

        self.add_settings_group(
            PlotSettingsMisfit(), MisfitPerRealView.Ids.PLOT_SETTINGS
        )
        self.add_settings_group(MisfitOptions(), MisfitPerRealView.Ids.MISFIT_OPTIONS)
        self.main_column = self.add_column(MisfitPerRealView.Ids.MAIN_COLUMN)

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.layout_element(MisfitPerRealView.Ids.MAIN_COLUMN)
                .get_unique_id()
                .to_string(),
                "children",
            ),
            Input(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_ENSEMBLES), "data"
            ),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_DATES), "data"),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_PHASE), "data"),
            Input(self.get_store_unique_id(PluginIds.Stores.SELECTED_WELLS), "data"),
            Input(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_WELL_COLLECTIONS),
                "data",
            ),
            Input(
                self.get_store_unique_id(
                    PluginIds.Stores.SELECTED_COMBINE_WELLS_COLLECTION
                ),
                "data",
            ),
            Input(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_REALIZATIONS), "data"
            ),
            Input(
                self.settings_group(MisfitPerRealView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsMisfit.Ids.COLORBY)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(MisfitPerRealView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsMisfit.Ids.SORTING_RANKING)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(MisfitPerRealView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsMisfit.Ids.FIG_LAYOUT_HEIGHT)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(MisfitPerRealView.Ids.MISFIT_OPTIONS)
                .component_unique_id(MisfitOptions.Ids.MISFIT_WEIGHT)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(MisfitPerRealView.Ids.MISFIT_OPTIONS)
                .component_unique_id(MisfitOptions.Ids.MISFIT_EXPONENT)
                .to_string(),
                "value",
            ),
        )
        def _update_plots(
            ensemble_names: List[str],
            selector_dates: list,
            selector_phases: list,
            selector_well_names: list,
            selector_well_collection_names: list,
            selector_well_combine_type: str,
            selector_realizations: List[int],
            colorby: str,
            sorting: str,
            figheight: int,
            obs_error_weight: float,
            misfit_exponent: float,
        ) -> Union[str, List[Component]]:

            if not ensemble_names:
                return "No ensembles selected"

            well_names = _get_well_names_combined(
                self.well_collections,
                selector_well_collection_names,
                selector_well_names,
                selector_well_combine_type,
            )

            dframe = makedf.get_df_diff(
                makedf.get_df_smry(
                    self.input_provider_set,
                    ensemble_names,
                    self.ens_vectors,
                    self.ens_realizations,
                    selector_realizations,
                    well_names,
                    selector_phases,
                    selector_dates,
                ),
                obs_error_weight,
                self.weight_reduction_factor_oil,
                self.weight_reduction_factor_wat,
                self.weight_reduction_factor_gas,
                misfit_exponent,
            )

            figures = makefigs.prod_misfit_plot(
                dframe,
                selector_phases,
                colorby,
                sorting,
                figheight,
                misfit_exponent,
                # misfit_normalization,
            )

            return figures



from typing import Callable, Dict, List, Optional

import webviz_core_components as wcc
from dash import html
from webviz_subsurface_components import LeafletMap


def intersection_and_map_layout(get_uuid: Callable) -> html.Div:
    """Layout for the intersection graph and maps"""
    return html.Div(
        children=[
            wcc.Frame(
                highlight=False,
                color="white",
                children=[
                    wcc.Graph(
                        style={"height": "48vh", "background-color": "white"},
                        id=get_uuid("intersection-graph"),
                    ),
                ],
            ),
            wcc.Frame(
                id=get_uuid("maps-and-uncertainty-table-wrapper"),
                highlight=False,
                color="white",
                children=[
                    wcc.FlexBox(
                        style={"height": "40vh", "display": "flex"},
                        id=get_uuid("all-maps-wrapper"),
                        children=[
                            html.Div(
                                style={"flex": 1},
                                id=get_uuid("map-wrapper"),
                                children=map_layout(
                                    uuid=get_uuid("map"),
                                    leaflet_id=get_uuid("leaflet-map1"),
                                    synced_uuids=[
                                        get_uuid("leaflet-map2"),
                                        get_uuid("leaflet-map3"),
                                    ],
                                    draw_polyline=True,
                                ),
                            ),
                            html.Div(
                                style={"flex": 1},
                                children=map_layout(
                                    uuid=get_uuid("map2"),
                                    leaflet_id=get_uuid("leaflet-map2"),
                                    synced_uuids=[
                                        get_uuid("leaflet-map1"),
                                        get_uuid("leaflet-map3"),
                                    ],
                                ),
                            ),
                            html.Div(
                                style={"flex": 1},
                                children=map_layout(
                                    uuid=get_uuid("map3"),
                                    leaflet_id=get_uuid("leaflet-map3"),
                                    synced_uuids=[
                                        get_uuid("leaflet-map1"),
                                        get_uuid("leaflet-map2"),
                                    ],
                                ),
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )


def map_layout(
    uuid: str,
    leaflet_id: str,
    synced_uuids: Optional[List[str]] = None,
    draw_polyline: bool = False,
) -> html.Div:
    synced_uuids = synced_uuids if synced_uuids else []
    props: Optional[Dict] = (
        {
            "drawTools": {
                "drawMarker": False,
                "drawPolygon": False,
                "drawPolyline": True,
                "position": "topright",
            }
        }
        if draw_polyline
        else {}
    )
    return html.Div(
        children=[
            html.Label(
                style={"textAlign": "center", "fontSize": "0.8em"},
                id={"id": uuid, "element": "label"},
            ),
            html.Div(
                style={
                    "height": "37vh",
                },
                children=LeafletMap(
                    syncedMaps=synced_uuids,
                    id=leaflet_id,
                    layers=[],
                    unitScale={},
                    autoScaleMap=True,
                    minZoom=-19,
                    updateMode="replace",
                    mouseCoords={"position": "bottomright"},
                    colorBar={"position": "bottomleft"},
                    switch={
                        "value": False,
                        "disabled": False,
                        "label": "Hillshading",
                    },
                    **props
                ),
            ),
        ],
    )



from typing import Optional

import pyarrow as pa

from .ensemble_summary_provider import VectorMetadata


def is_rate_from_field_meta(field: pa.Field) -> bool:
    """Determine if the field is a rate by querying for the "is_rate" keyword in the
    field's metadata.
    Will silently return False if no metadata exists for the field, but will raise an
    exception if the field HAS metadata but the 'is_rate' key is missing.
    """
    # What should be done here if the is_rate key is not found?
    # Should we just assume that it is not a rate or throw an exception?

    meta_dict = field.metadata
    if meta_dict is not None:
        try:
            is_rate_bytestr = meta_dict[b"is_rate"]
            return bool(is_rate_bytestr == b"True")
        except KeyError as exc:
            raise KeyError(
                f"Field {field.name} has metadata, but the is_rate key was not found"
            ) from exc
    return False

    # meta_dict = field.metadata
    # if meta_dict is None:
    #     raise ValueError(f"Field {field.name} has no metadata")
    # is_rate_bytestr = meta_dict.get(b"is_rate")
    # if is_rate_bytestr is None:
    #     raise KeyError(f"is_rate key not found in metadata for field {field.name}")
    # return bool(is_rate_bytestr == b"True")

    # try:
    #     return bool(field.metadata[b"is_rate"] == b"True")
    # except (TypeError, KeyError) as e:
    #     raise ValueError(
    #         "Field is missing metadata or 'is_rate' key was not found"
    #     ) from e


def create_vector_metadata_from_field_meta(
    field: pa.Field,
) -> Optional[VectorMetadata]:
    """Create VectorMetadata from keywords stored in the field's metadata"""

    # Note that when we query the values in the field.metadata we always get byte strings
    # back from pyarrow. Further, ecl2df writes all the values as strings, so we must
    # convert these to the correct types before creating the VectorMetadata instance.
    # See also ecl2df code:
    # https://github.com/equinor/ecl2df/blob/0e30fb8046bf17fd338bb468584985c5d816e2f6/ecl2df/summary.py#L441

    # Currently, based on the ecl2df code, we assume that all keys except for 'get_num'
    # and 'wgname' must be present in order to return a valid metadata object
    # https://github.com/equinor/ecl2df/blob/0e30fb8046bf17fd338bb468584985c5d816e2f6/ecl2df/summary.py#L541-L552

    meta_dict = field.metadata
    if not meta_dict:
        return None

    try:
        unit_bytestr = meta_dict[b"unit"]
        is_total_bytestr = meta_dict[b"is_total"]
        is_rate_bytestr = meta_dict[b"is_rate"]
        is_historical_bytestr = meta_dict[b"is_historical"]
        keyword_bytestr = meta_dict[b"keyword"]
    except KeyError:
        return None

    wgname_bytestr = meta_dict.get(b"wgname")
    get_num_bytestr = meta_dict.get(b"get_num")

    # print(f"unit_bytestr:  |{unit_bytestr}|")
    # print(f"is_total_bytestr:  |{is_total_bytestr}|")
    # print(f"is_rate_bytestr:  |{is_rate_bytestr}|")
    # print(f"is_historical_bytestr:  |{is_historical_bytestr}|")
    # print(f"keyword_bytestr:  |{keyword_bytestr}|")
    # print(f"wgname_bytestr:  |{wgname_bytestr}|")
    # print(f"get_num_bytestr:  |{get_num_bytestr}|")

    wgname: Optional[str] = None
    if wgname_bytestr and wgname_bytestr != b"None":
        wgname = wgname_bytestr.decode("ascii")

    get_num: Optional[int] = None
    if get_num_bytestr and get_num_bytestr != b"None":
        get_num = int(get_num_bytestr.decode("ascii"))

    return VectorMetadata(
        unit=unit_bytestr.decode("ascii"),
        is_total=bool(is_total_bytestr == b"True"),
        is_rate=bool(is_rate_bytestr == b"True"),
        is_historical=bool(is_historical_bytestr == b"True"),
        keyword=keyword_bytestr.decode("ascii"),
        wgname=wgname,
        get_num=get_num,
    )



import webviz_core_components as wcc
from dash import html
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import ViewElementABC


class WellOverviewViewElement(ViewElementABC):
    class Ids(StrEnum):
        GRAPH = "graph"

    def __init__(self) -> None:
        super().__init__()

    def inner_layout(self) -> html.Div:
        return html.Div(
            children=wcc.Graph(
                id=self.register_component_unique_id(WellOverviewViewElement.Ids.GRAPH),
                style={"height": "87vh"},
            ),
        )



import datetime

import pandas as pd

from webviz_subsurface._utils.dataframe_utils import (
    assert_date_column_is_datetime_object,
    make_date_column_datetime_object,
)


def create_relative_to_date_df(
    df: pd.DataFrame, relative_date: datetime.datetime
) -> pd.DataFrame:
    """
    Create dataframe where data for relative_date is subtracted from respective
    vector data.

    I.e. Subtract realization data at given relative date from corresponding
    realizations at each individual date for each vector column in dataframe.

    `Assume:`
    Set of realizations are equal for each date in "DATE" column of input dataframe.

    `Input:`
    * df - `Columns` in dataframe: ["DATE", "REAL", vector1, ..., vectorN]

    `Output:`
    * df - `Columns` in dataframe: ["DATE", "REAL", vector1, ..., vectorN]

    NOTE:
    - This function iterates over realization group in input dataframe
    - For-loop makes it possible to get realization not present in relative_date_df, if
    realization is not present in relative_date_df the realization is excluded output.
    """

    assert_date_column_is_datetime_object(df)

    if not set(["DATE", "REAL"]).issubset(set(df.columns)):
        raise ValueError('Expect column "DATE" and "REAL" in input dataframe!')

    # Columns of correct dtype
    _columns = {name: pd.Series(dtype=df.dtypes[name]) for name in df.columns}
    output_df = pd.DataFrame(_columns)

    relative_date_df: pd.DataFrame = df.loc[df["DATE"] == relative_date].drop(
        columns=["DATE"]
    )
    if relative_date_df.empty:
        # Dataframe with columns, but no rows
        return output_df

    vectors = [elm for elm in df.columns if elm not in ("DATE", "REAL")]

    # NOTE: This for-loop makes it possible to get real not represented in relative_date_df!
    for real, real_df in df.groupby("REAL"):
        relative_date_data = relative_date_df.loc[
            relative_date_df["REAL"] == real
        ].drop(columns=["REAL"])

        # If realization does not exist in relative_date_df
        if relative_date_data.empty:
            continue

        real_df[vectors] = real_df[vectors].sub(relative_date_data.iloc[0], axis=1)
        output_df = pd.concat([output_df, real_df], ignore_index=True)

    make_date_column_datetime_object(output_df)
    return output_df



import hashlib
import logging
import os
from pathlib import Path
from typing import List

from webviz_config.webviz_factory import WebvizFactory
from webviz_config.webviz_factory_registry import WEBVIZ_FACTORY_REGISTRY
from webviz_config.webviz_instance_info import WebvizRunMode

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._egrid_file_discovery import discover_per_realization_eclipse_files
from ._roff_file_discovery import discover_per_realization_roff_files
from .ensemble_grid_provider import EnsembleGridProvider
from .provider_impl_egrid import ProviderImplEgrid
from .provider_impl_roff import ProviderImplRoff

LOGGER = logging.getLogger(__name__)


class EnsembleGridProviderFactory(WebvizFactory):
    def __init__(
        self,
        root_storage_folder: Path,
        allow_storage_writes: bool,
        avoid_copying_grid_data: bool,
    ) -> None:
        self._storage_dir = Path(root_storage_folder) / __name__
        self._allow_storage_writes = allow_storage_writes
        self._avoid_copying_grid_data = avoid_copying_grid_data

        LOGGER.info(
            f"EnsembleGridProviderFactory init: storage_dir={self._storage_dir}"
        )

        if self._allow_storage_writes:
            os.makedirs(self._storage_dir, exist_ok=True)

    @staticmethod
    def instance() -> "EnsembleGridProviderFactory":
        """Static method to access the singleton instance of the factory."""

        factory = WEBVIZ_FACTORY_REGISTRY.get_factory(EnsembleGridProviderFactory)
        if not factory:
            app_instance_info = WEBVIZ_FACTORY_REGISTRY.app_instance_info
            storage_folder = app_instance_info.storage_folder
            allow_writes = app_instance_info.run_mode != WebvizRunMode.PORTABLE
            dont_copy_grid_data = (
                app_instance_info.run_mode == WebvizRunMode.NON_PORTABLE
            )

            factory = EnsembleGridProviderFactory(
                root_storage_folder=storage_folder,
                allow_storage_writes=allow_writes,
                avoid_copying_grid_data=dont_copy_grid_data,
            )

            # Store the factory object in the global factory registry
            WEBVIZ_FACTORY_REGISTRY.set_factory(EnsembleGridProviderFactory, factory)

        return factory

    def create_from_roff_files(
        self, ens_path: str, grid_name: str, attribute_filter: List[str] = None
    ) -> EnsembleGridProvider:
        timer = PerfTimer()
        string_to_hash = (
            f"{ens_path}_{grid_name}"
            if attribute_filter is None
            else f"{ens_path}_{grid_name}_{'_'.join([str(attr) for attr in attribute_filter])}"
        )
        storage_key = f"ens__{_make_hash_string(string_to_hash)}"
        provider = ProviderImplRoff.from_backing_store(self._storage_dir, storage_key)
        if provider:
            LOGGER.info(
                f"Loaded grid provider from backing store in {timer.elapsed_s():.2f}s ("
                f"ens_path={ens_path})"
            )
            return provider

        # We can only import data from data source if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(f"Failed to load grid provider for {ens_path}")

        LOGGER.info(f"Importing/copying grid data for: {ens_path}")

        timer.lap_s()
        grid_info, grid_parameters_info = discover_per_realization_roff_files(
            ens_path, grid_name, attribute_filter
        )

        # As an optimization, avoid copying the grid data into the backing store,
        # typically when  we're running in non-portable mode
        ProviderImplRoff.write_backing_store(
            self._storage_dir,
            storage_key,
            grid_geometries_info=grid_info,
            grid_parameters_info=grid_parameters_info,
            avoid_copying_grid_data=self._avoid_copying_grid_data,
        )
        et_write_s = timer.lap_s()

        provider = ProviderImplRoff.from_backing_store(self._storage_dir, storage_key)
        if not provider:
            raise ValueError(f"Failed to load/create grid provider for {ens_path}")

        LOGGER.info(
            f"Saved grid provider to backing store in {timer.elapsed_s():.2f}s ("
            f" write={et_write_s:.2f}s, ens_path={ens_path})"
        )

        return provider

    def create_from_eclipse_files(
        self,
        ens_path: str,
        grid_name: str,
        init_properties: List[str],
        restart_properties: List[str],
    ) -> EnsembleGridProvider:
        timer = PerfTimer()

        string_to_hash = f"{ens_path}_{grid_name}_egrid"
        storage_key = f"ens__{_make_hash_string(string_to_hash)}"
        provider = ProviderImplEgrid.from_backing_store(
            self._storage_dir, storage_key, init_properties, restart_properties
        )
        if provider:
            LOGGER.info(
                f"Loaded grid provider from backing store in {timer.elapsed_s():.2f}s ("
                f"ens_path={ens_path})"
            )
            return provider

        # We can only import data from data source if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(f"Failed to load grid provider for {ens_path}")

        LOGGER.info(f"Importing/copying grid data for: {ens_path}")

        timer.lap_s()
        eclipse_case_paths = discover_per_realization_eclipse_files(ens_path, grid_name)

        # As an optimization, avoid copying the grid data into the backing store,
        # typically when  we're running in non-portable mode
        ProviderImplEgrid.write_backing_store(
            self._storage_dir,
            storage_key,
            eclipse_case_paths=eclipse_case_paths,
            avoid_copying_grid_data=self._avoid_copying_grid_data,
        )
        et_write_s = timer.lap_s()
        provider = ProviderImplEgrid.from_backing_store(
            self._storage_dir, storage_key, init_properties, restart_properties
        )
        if not provider:
            raise ValueError(f"Failed to load/create grid provider for {ens_path}")

        LOGGER.info(
            f"Saved grid provider to backing store in {timer.elapsed_s():.2f}s ("
            f" write={et_write_s:.2f}s, ens_path={ens_path})"
        )

        return provider


def _make_hash_string(string_to_hash: str) -> str:
    # There is no security risk here and chances of collision should be very slim
    return hashlib.md5(string_to_hash.encode()).hexdigest()  # nosec



import pathlib
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import pandas as pd

from .ensemble_model import EnsembleModel


class EnsembleSetModel:
    """Class to load and manipulate ensemble sets from given paths to
    ensembles on disk"""

    def __init__(
        self,
        ensemble_paths: dict,
        smry_time_index: Optional[Union[list, str]] = None,
        smry_column_keys: Optional[list] = None,
    ) -> None:
        self._ensemble_paths = ensemble_paths
        self._webvizstore: List = []
        self._ensembles = [
            EnsembleModel(ens_name, ens_path, filter_file="OK")
            for ens_name, ens_path in self._ensemble_paths.items()
        ]

        self._smry_time_index = smry_time_index
        self._smry_column_keys = smry_column_keys
        self._cached_smry_df: Optional[pd.DataFrame] = None
        self._hash_for_cached_smry_df: Optional[pd.Series] = None

    def __repr__(self) -> str:
        return f"EnsembleSetModel: {self._ensemble_paths}"

    @property
    def ens_folders(self) -> dict:
        """Get root folders for ensemble set"""
        return {
            ens: pathlib.Path(ens_path.split("realization")[0])
            for ens, ens_path in self._ensemble_paths.items()
        }

    @staticmethod
    def _get_ensembles_data(
        ensemble_models: List[EnsembleModel], func: str, **kwargs: Any
    ) -> pd.DataFrame:
        """Runs the provided function for each ensemble and concats dataframes"""
        dfs = []
        for ensemble in ensemble_models:
            try:
                dframe = getattr(ensemble, func)(**kwargs)
                dframe.insert(0, "ENSEMBLE", ensemble.ensemble_name)
                dfs.append(dframe)
            except (KeyError, ValueError):
                # Happens if an ensemble is missing some data
                # Warning has already been issued at initialization
                pass
        if dfs:
            return pd.concat(dfs, sort=False)
        raise KeyError(f"No data found for {func} with arguments: {kwargs}")

    def load_parameters(self) -> pd.DataFrame:
        return EnsembleSetModel._get_ensembles_data(self._ensembles, "load_parameters")

    def get_or_load_smry_cached(self) -> pd.DataFrame:
        """Either loads smry data from file or retrieves a cached copy of DataFrame.
        Note that it is imperative that the returned DataFrame be treated as read-only
        since it will probably be shared by multiple clients.
        """

        if self._cached_smry_df is not None:
            # If we're returning cached data frame, try and verify that it hasn't been tampered with
            curr_hash: pd.Series = pd.util.hash_pandas_object(self._cached_smry_df)
            if not curr_hash.equals(self._hash_for_cached_smry_df):
                raise KeyError("The cached SMRY DataFrame has been tampered with")

            return self._cached_smry_df

        self._cached_smry_df = EnsembleSetModel._get_ensembles_data(
            self._ensembles,
            "load_smry",
            time_index=self._smry_time_index,
            column_keys=self._smry_column_keys,
        )
        self._hash_for_cached_smry_df = pd.util.hash_pandas_object(self._cached_smry_df)

        return self._cached_smry_df

    def load_smry_meta(self) -> pd.DataFrame:
        """Finds metadata for the summary vectors in the ensemble set.
        Note that we assume the same units for all ensembles.
        (meaning that we update/overwrite when checking the next ensemble)
        """

        smry_meta: dict = {}
        for ensemble in self._ensembles:
            smry_meta.update(
                ensemble.load_smry_meta(column_keys=self._smry_column_keys).T.to_dict()
            )
        return pd.DataFrame(smry_meta).transpose()

    def load_csv(self, csv_file: pathlib.Path) -> pd.DataFrame:
        return EnsembleSetModel._get_ensembles_data(
            self._ensembles, "load_csv", csv_file=csv_file
        )

    @property
    def webvizstore(self) -> List[Tuple[Callable, List[Dict]]]:
        store_functions = []
        for ensemble in self._ensembles:
            store_functions.extend(ensemble.webviz_store)
        return store_functions



import logging

import numpy as np
import xtgeo
from vtkmodules.util.numpy_support import (
    numpy_to_vtk,
    numpy_to_vtkIdTypeArray,
    vtk_to_numpy,
)

# pylint: disable=no-name-in-module,
from vtkmodules.vtkCommonCore import vtkPoints

# pylint: disable=no-name-in-module,
from vtkmodules.vtkCommonDataModel import (
    vtkCellArray,
    vtkDataSetAttributes,
    vtkExplicitStructuredGrid,
)

from webviz_subsurface._utils.perf_timer import PerfTimer

LOGGER = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
def xtgeo_grid_to_vtk_explicit_structured_grid(
    xtg_grid: xtgeo.Grid,
) -> vtkExplicitStructuredGrid:

    timer = PerfTimer()

    # Create geometry data suitable for use with VTK's explicit structured grid
    # based on the specified xtgeo 3d grid
    pt_dims, vertex_arr, conn_arr, inactive_arr = xtg_grid.get_vtk_esg_geometry_data()
    vertex_arr[:, 2] *= -1
    et_get_esg_geo_data_ms = timer.lap_ms()

    # LOGGER.debug(f"pt_dims={pt_dims}")
    # LOGGER.debug(f"vertex_arr.shape={vertex_arr.shape}")
    # LOGGER.debug(f"vertex_arr.dtype={vertex_arr.dtype}")
    # LOGGER.debug(f"conn_arr.shape={conn_arr.shape}")
    # LOGGER.debug(f"conn_arr.dtype={conn_arr.dtype}")

    vtk_esgrid = _create_vtk_esgrid_from_verts_and_conn(pt_dims, vertex_arr, conn_arr)
    et_create_vtk_esg_ms = timer.lap_ms()

    # Make sure we hide the inactive cells.
    # First we let VTK allocate cell ghost array, then we obtain a numpy view
    # on the array and write to that (we're actually modifying the native VTK array)
    ghost_arr_vtk = vtk_esgrid.AllocateCellGhostArray()
    ghost_arr_np = vtk_to_numpy(ghost_arr_vtk)
    ghost_arr_np[inactive_arr] = vtkDataSetAttributes.HIDDENCELL

    LOGGER.debug(
        f"xtgeo_grid_to_vtk_explicit_structured_grid() took {timer.elapsed_s():.2f}s "
        f"(get_esg_geo_data={et_get_esg_geo_data_ms}ms, "
        f"create_vtk_esg={et_create_vtk_esg_ms}ms)"
    )

    return vtk_esgrid


# -----------------------------------------------------------------------------
def _create_vtk_esgrid_from_verts_and_conn(
    point_dims: np.ndarray, vertex_arr_np: np.ndarray, conn_arr_np: np.ndarray
) -> vtkExplicitStructuredGrid:

    vertex_arr_np = vertex_arr_np.reshape(-1, 3)
    points_vtkarr = numpy_to_vtk(vertex_arr_np, deep=1)
    vtk_points = vtkPoints()
    vtk_points.SetData(points_vtkarr)

    # conn_idarr = numpy_to_vtk(conn_arr_np, deep=1, array_type=vtkConstants.VTK_ID_TYPE)
    conn_idarr = numpy_to_vtkIdTypeArray(conn_arr_np, deep=1)
    vtk_cell_array = vtkCellArray()
    vtk_cell_array.SetData(8, conn_idarr)

    vtk_esgrid = vtkExplicitStructuredGrid()
    vtk_esgrid.SetDimensions(point_dims)
    vtk_esgrid.SetPoints(vtk_points)
    vtk_esgrid.SetCells(vtk_cell_array)

    vtk_esgrid.ComputeFacesConnectivityFlagsArray()

    return vtk_esgrid



import datetime
import shutil
import warnings
from pathlib import Path
from typing import Callable, List, Optional, Tuple

import pandas as pd
import webviz_core_components as wcc
from dash import html
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore


class DiskUsage(WebvizPluginABC):
    """Visualize disk usage in a FMU project. It adds a dashboard showing disk usage per user.

    ---

    * **`scratch_dir`:** Path to the scratch directory to show disk usage for.
    * **`date`:** Date as string of form YYYY-MM-DD to request an explisit date. Default is to
    to use the most recent file avaialable, limited to the last week.

    ---

    ?> The `scratch_dir` directory must have a hidden folder `.disk_usage` containing daily
    csv files called `disk_usage_user_YYYY-MM-DD.csv`, where YYYY-MM-DD is the date.
    The plugin will search backwards from the current date, and throw an error if no file was found
    from the last week.

    The csv file must have the columns `userid` and `usageKB` (where KB means
    [kibibytes](https://en.wikipedia.org/wiki/Kibibyte)). All other columns are ignored.
    """

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        scratch_dir: Path,
        date: Optional["str"] = None,
    ):

        super().__init__()

        self.scratch_dir = scratch_dir
        self.date_input = date
        self.disk_usage = get_disk_usage(self.scratch_dir, self.date_input)
        self.date = str(self.disk_usage["date"][0])
        self.theme = webviz_settings.theme

    @property
    def layout(self) -> html.Div:
        return html.Div(
            [
                wcc.Header(
                    f"Disk usage on {self.scratch_dir} per user as of {self.date}",
                    style={"text-align": "center"},
                ),
                wcc.FlexBox(
                    children=[
                        wcc.Frame(
                            color="white",
                            children=wcc.FlexColumn(
                                wcc.Graph(
                                    figure=self.pie_chart,
                                    config={"displayModeBar": False},
                                )
                            ),
                        ),
                        wcc.Frame(
                            color="white",
                            children=wcc.FlexColumn(
                                flex=2,
                                children=wcc.Graph(figure=self.bar_chart),
                            ),
                        ),
                    ]
                ),
            ]
        )

    @property
    def pie_chart(self) -> dict:
        return {
            "data": [
                {
                    "values": self.disk_usage["usageGiB"],
                    "labels": self.disk_usage["userid"],
                    "pull": (self.disk_usage["userid"].values == "<b>Free space</b>")
                    * 0.05,
                    "text": self.disk_usage["usageGiB"].map(lambda x: f"{x:.2f} GiB"),
                    "textinfo": "label",
                    "textposition": "inside",
                    "hoverinfo": "label+text",
                    "type": "pie",
                }
            ],
            "layout": self.theme.create_themed_layout({}),
        }

    @property
    def bar_chart(self) -> dict:
        return {
            "data": [
                {
                    "y": self.disk_usage["usageGiB"],
                    "x": self.disk_usage["userid"],
                    "text": self.disk_usage["usageGiB"].map(lambda x: f"{x:.2f} GiB"),
                    "hoverinfo": "x+text",
                    "type": "bar",
                }
            ],
            "layout": self.theme.create_themed_layout(
                {"yaxis": {"title": "GiB (GibiBytes)"}}
            ),
        }

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [
            (
                get_disk_usage,
                [{"scratch_dir": self.scratch_dir, "date": self.date_input}],
            )
        ]


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def get_disk_usage(scratch_dir: Path, date: Optional[str]) -> pd.DataFrame:
    if date is None:
        df, date = _loop_dates(scratch_dir)
    else:
        df = _get_disk_usage_for_date(scratch_dir, date)
        if df is None:
            raise FileNotFoundError(
                f"No disk usage file found for {date} in {scratch_dir}."
            )

    df.rename(
        columns={"usageKB": "usageKiB"}, inplace=True
    )  # Old format had an error (KB instead of KiB)

    df["usageGiB"] = df["usageKiB"] / (1024**2)

    df.drop(columns="usageKiB", inplace=True)

    free_space_gib = _estimate_free_space(df, scratch_dir, date)
    if free_space_gib < 0:
        warnings.warn(
            f"Reported disk usage in "
            f"{scratch_dir}/.disk_usage for {date}"
            f"must be wrong (unless total disk size has changed after {date}). "
            f"Total reported usage is {int(df['usageGiB'])} GiB, "
            f"but the disk size is only {int(shutil.disk_usage(scratch_dir).total / 1024**3)} GiB "
            f"(this would imply negative free space of {int(free_space_gib)} GiB).",
            UserWarning,
            stacklevel=0,
        )
    else:
        df = pd.concat(
            [
                df,
                pd.DataFrame.from_records(
                    [
                        {
                            "userid": "<b>Free space</b>",
                            "usageGiB": free_space_gib,
                        }
                    ]
                ),
            ],
            ignore_index=True,
        )

    df["date"] = date
    return df.sort_values(by="usageGiB", axis=0, ascending=False)


def _get_disk_usage_for_date(scratch_dir: Path, date: str) -> Optional[pd.DataFrame]:
    csv_file = scratch_dir / ".disk_usage" / f"disk_usage_user_test_{date}.csv"
    if csv_file.exists():
        return pd.read_csv(csv_file)
    # If file does not exist, look for old format
    csv_file = scratch_dir / ".disk_usage" / f"disk_usage_user_{date}.csv"
    if csv_file.exists():
        return pd.read_csv(csv_file)
    return None


def _loop_dates(scratch_dir: Path) -> Tuple[pd.DataFrame, str]:
    today = datetime.datetime.today()
    for i in range(7):
        date = (today - datetime.timedelta(days=i)).strftime("%Y-%m-%d")
        df = _get_disk_usage_for_date(scratch_dir, date)
        if df is not None:
            return (df, date)
    raise FileNotFoundError(
        f"No disk usage file found for the last week in {scratch_dir}."
    )


def _estimate_free_space(df: pd.DataFrame, scratch_dir: Path, date: str) -> float:
    txt_file = Path(scratch_dir / ".disk_usage" / f"disk_usage_user_test_{date}.txt")
    total_usage = None
    if txt_file.exists():
        lines = txt_file.read_text().splitlines()
        for line in lines[::-1]:  # reversed as last line is most likely
            if line.startswith("Total usage without overhead"):
                line = line[len("Total usage without overhead") :]
                total_usage = float(line.split()[0])
                break

    total_usage = df["usageGiB"].sum() if total_usage is None else total_usage
    return shutil.disk_usage(scratch_dir).total / (1024**3) - total_usage



import abc
from dataclasses import dataclass
from typing import List, Optional, Sequence

import pandas as pd


@dataclass(frozen=True)
class ColumnMetadata:
    unit: Optional[str]


class EnsembleTableProvider(abc.ABC):
    @abc.abstractmethod
    def column_names(self) -> List[str]:
        ...

    @abc.abstractmethod
    def realizations(self) -> List[int]:
        ...

    @abc.abstractmethod
    def get_column_data(
        self, column_names: Sequence[str], realizations: Optional[Sequence[int]] = None
    ) -> pd.DataFrame:
        ...

    @abc.abstractmethod
    def column_metadata(self, column_name: str) -> Optional[ColumnMetadata]:
        """Returns metadata for the specified column.

        Returns None if no metadata is found for the column.
        Returns a empty ColumnMetadata object if there is metadata, but it's
        not the columns specified in ColumnMetadata.
        """
